@@ -29,6 +29,7 @@
   #:use-module (srfi srfi-34)
   #:use-module (srfi srfi-35)
   #:use-module (rnrs io ports)
+  #:use-module ((ice-9 control) #:select (call/ec))
   #:use-module (ice-9 match)
   #:use-module (ice-9 format)
   #:autoload   (ice-9 pretty-print) (truncated-print)
@@ -76,6 +77,7 @@
             make-kill-destructor
             exec-command
             fork+exec-command
+            default-pid-file-timeout
             read-pid-file
             make-system-constructor
             make-system-destructor
@@ -719,7 +721,11 @@ results."
   ;; set when starting a service.
   (make-parameter (environ)))
 
-(define* (read-pid-file file #:key (max-delay 5)
+(define default-pid-file-timeout
+  ;; Maximum number of seconds to wait for a PID file to show up.
+  (make-parameter 5))
+
+(define* (read-pid-file file #:key (max-delay (default-pid-file-timeout))
                         (validate-pid? #f))
   "Wait for MAX-DELAY seconds for FILE to show up, and read its content as a
 number.  Return #f if FILE was not created or does not contain a number;
@@ -771,12 +777,14 @@ daemon writing FILE is running in a separate PID namespace."
                        (group #f)
                        (log-file #f)
                        (directory (default-service-directory))
+                       (file-creation-mask #f)
                        (environment-variables (default-environment-variables)))
-  "Run COMMAND as the current process from DIRECTORY, and with
-ENVIRONMENT-VARIABLES (a list of strings like \"PATH=/bin\".)  File
-descriptors 1 and 2 are kept as is or redirected to LOG-FILE if it's true,
-whereas file descriptor 0 (standard input) points to /dev/null; all other file
-descriptors are closed prior to yielding control to COMMAND.
+  "Run COMMAND as the current process from DIRECTORY, with FILE-CREATION-MASK
+if it's true, and with ENVIRONMENT-VARIABLES (a list of strings like
+\"PATH=/bin\").  File descriptors 1 and 2 are kept as is or redirected to
+LOG-FILE if it's true, whereas file descriptor 0 (standard input) points to
+/dev/null; all other file descriptors are closed prior to yielding control to
+COMMAND.
 
 By default, COMMAND is run as the current user.  If the USER keyword
 argument is present and not false, change to USER immediately before
@@ -808,7 +816,7 @@ false."
              ;; Redirect stout and stderr to use LOG-FILE.
              (catch-system-error (close-fdes 1))
              (catch-system-error (close-fdes 2))
-             (dup2 (open-fdes log-file (logior O_CREAT O_WRONLY O_APPEND)) 1)
+             (dup2 (open-fdes log-file (logior O_CREAT O_WRONLY O_APPEND) #o640) 1)
              (dup2 1 2))
            (lambda (key . args)
              (format (current-error-port)
@@ -840,30 +848,26 @@ false."
            (print-exception (current-error-port) #f key args)
            (primitive-exit 1))))
 
-     ;; As the last action, close file descriptors.  Doing it last makes
-     ;; "error in the finalization thread: Bad file descriptor" issues
-     ;; unlikely on 2.2.
-     (let loop ((i 3))
-       (when (< i max-fd)
-         ;; First try to close any ports associated with file descriptor I.
-         ;; Otherwise the finalization thread might get around to closing
-         ;; those ports eventually, which will raise an EBADF exception (on
-         ;; 2.2), leading to messages like "error in the finalization
-         ;; thread: Bad file descriptor".
-         (for-each (lambda (port)
-                     (catch-system-error (close-port port)))
-                   (fdes->ports i))
-         (catch-system-error (close-fdes i))
-         (loop (+ i 1)))))
+     (when file-creation-mask
+       (umask file-creation-mask))
 
-     (catch 'system-error
-       (lambda ()
-         (apply execlp program program args))
-       (lambda args
-         (format (current-error-port)
-                 "exec of ~s failed: ~a~%"
-                 program (strerror (system-error-errno args)))
-         (primitive-exit 1))))))
+     ;; Last, close all file descriptors.  Do that after shutting down the
+     ;; finalization thread since we will close its pipe, leading to
+     ;; "error in the finalization thread: Bad file descriptor".
+     (without-automatic-finalization
+      (let loop ((i 3))
+        (when (< i max-fd)
+          (catch-system-error (close-fdes i))
+          (loop (+ i 1))))
+
+      (catch 'system-error
+        (lambda ()
+          (apply execlp program program args))
+        (lambda args
+          (format (current-error-port)
+                  "exec of ~s failed: ~a~%"
+                  program (strerror (system-error-errno args)))
+          (primitive-exit 1))))))))
 
 (define* (fork+exec-command command
                             #:key
@@ -871,11 +875,12 @@ false."
                             (group #f)
                             (log-file #f)
                             (directory (default-service-directory))
+                            (file-creation-mask #f)
                             (environment-variables
                              (default-environment-variables)))
   "Spawn a process that executed COMMAND as per 'exec-command', and return
 its PID."
-  ;; Install the SIGCHLD handler if this is the first fork+exec-command call
+  ;; Install the SIGCHLD handler if this is the first fork+exec-command call.
   (unless %sigchld-handler-installed?
     (sigaction SIGCHLD handle-SIGCHLD SA_NOCLDSTOP)
     (set! %sigchld-handler-installed? #t))
@@ -886,24 +891,26 @@ its PID."
                       #:group group
                       #:log-file log-file
                       #:directory directory
+                      #:file-creation-mask file-creation-mask
                       #:environment-variables environment-variables)
         pid)))
 
-(define %pid-file-timeout
-  ;; Maximum number of seconds we wait for a PID file to show up.
-  5)
-
-(define make-forkexec-constructor
-  (let ((warn-deprecated-form
-         ;; Until 0.1, this procedure took a rest list.
-         (lambda ()
-           (issue-deprecation-warning
-            "This 'make-forkexec-constructor' form is deprecated; use
- (make-forkexec-constructor '(\"PROGRAM\" \"ARGS\"...)."))))
-    (case-lambda*
-     "Return a procedure that forks a child process, closes all file
+(define* (make-forkexec-constructor command
+                                    #:key
+                                    (user #f)
+                                    (group #f)
+                                    (directory (default-service-directory))
+                                    (environment-variables
+                                     (default-environment-variables))
+                                    (file-creation-mask #f)
+                                    (pid-file #f)
+                                    (pid-file-timeout
+                                     (default-pid-file-timeout))
+                                    (log-file #f))
+  "Return a procedure that forks a child process, closes all file
 descriptors except the standard output and standard error descriptors, sets
-the current directory to @var{directory}, changes the environment to
+the current directory to @var{directory}, sets the umask to
+@var{file-creation-mask} unless it is @code{#f}, changes the environment to
 @var{environment-variables} (using the @code{environ} procedure), sets the
 current user to @var{user} and the current group to @var{group} unless they
 are @code{#f}, and executes @var{command} (a list of strings.)  The result of
@@ -914,60 +921,48 @@ the process being launched; the return value is the PID read from that file,
 once that file has been created.  If @var{pid-file} does not show up in less
 than @var{pid-file-timeout} seconds, the service is considered as failing to
 start."
-     ((command #:key
-               (user #f)
-               (group #f)
-               (directory (default-service-directory))
-               (environment-variables (default-environment-variables))
-               (pid-file #f)
-               (pid-file-timeout %pid-file-timeout)
-               (log-file #f))
-      (let ((command (if (string? command)
-                         (begin
-                           (warn-deprecated-form)
-                           (list command))
-                         command)))
-        (lambda args
-          (define (clean-up file)
-            (when file
-              (catch 'system-error
-                (lambda ()
-                  (delete-file file))
-                (lambda args
-                  (unless (= ENOENT (system-error-errno args))
-                    (apply throw args))))))
+  (lambda args
+    (define (clean-up file)
+      (when file
+        (catch 'system-error
+          (lambda ()
+            (delete-file file))
+          (lambda args
+            (unless (= ENOENT (system-error-errno args))
+              (apply throw args))))))
 
-          (clean-up pid-file)
+    (clean-up pid-file)
 
-          (let ((pid (fork+exec-command command
-                                        #:user user
-                                        #:group group
-                                        #:log-file log-file
-                                        #:directory directory
-                                        #:environment-variables
-                                        environment-variables)))
-            (if pid-file
-                (match (read-pid-file pid-file
-                                      #:max-delay pid-file-timeout
-                                      #:validate-pid? #t)
-                  (#f
-                   (catch-system-error (kill pid SIGTERM))
-                   #f)
-                  ((? integer? pid)
-                   pid))
-                pid)))))
-     ((program . program-args)
-      ;; The old form, documented until 0.1 included.
-      (warn-deprecated-form)
-      (make-forkexec-constructor (cons program program-args))))))
+    (let ((pid (fork+exec-command command
+                                  #:user user
+                                  #:group group
+                                  #:log-file log-file
+                                  #:directory directory
+                                  #:file-creation-mask file-creation-mask
+                                  #:environment-variables
+                                  environment-variables)))
+      (if pid-file
+          (match (read-pid-file pid-file
+                                #:max-delay pid-file-timeout
+                                #:validate-pid? #t)
+            (#f
+             ;; Send SIGTERM to the whole process group.
+             (catch-system-error (kill (- pid) SIGTERM))
+             #f)
+            ((? integer? pid)
+             pid))
+          pid))))
 
-;; Produce a destructor that sends SIGNAL to the process with the pid
-;; given as argument, where SIGNAL defaults to `SIGTERM'.
-(define make-kill-destructor
-  (lambda* (#:optional (signal SIGTERM))
-    (lambda (pid . args)
-      (kill pid signal)
-      #f)))
+(define* (make-kill-destructor #:optional (signal SIGTERM))
+  "Return a procedure that sends SIGNAL to the process group of the PID given
+as argument, where SIGNAL defaults to `SIGTERM'."
+  (lambda (pid . args)
+    ;; Kill the whole process group PID belongs to.  Don't assume that PID
+    ;; is a process group ID: that's not the case when using #:pid-file,
+    ;; where the process group ID is the PID of the process that
+    ;; "daemonized".
+    (kill (- (getpgid pid)) signal)
+    #f))
 
 ;; Produce a constructor that executes a command.
 (define (make-system-constructor . command)
@@ -1068,15 +1063,11 @@ returned in unspecified."
       (lambda ()
         (waitpid what flags))
       (lambda args
-        ;; Did we get ECHILD or something?  If we did, that's a problem,
-        ;; because this procedure is supposed to be called only upon
-        ;; SIGCHLD.
-        (let ((errno (system-error-errno args)))
-          (local-output (l10n "warning: 'waitpid' ~a failed unexpectedly: ~a")
-                        what (strerror errno))
-          '(0 . #f))))))
+        (if (memv (system-error-errno args) (list ECHILD EINTR))
+            '(0 . #f)
+            (apply throw args))))))
 
-(define (handle-SIGCHLD signum)
+(define* (handle-SIGCHLD #:optional (signum SIGCHLD))
   "Handle SIGCHLD, possibly by respawning the service that just died, or
 otherwise by updating its state."
   (let loop ()
@@ -1273,9 +1264,8 @@ file when persistence is enabled."
 unexpectedly stopped running. This procedure is used as a fallback on systems
 where prctl/PR_SET_CHILD_SUBREAPER is unsupported."
   (define (process-exists? pid)
-    (catch #t
-      (lambda () (kill pid 0) #t)
-      (lambda _ #f)))
+    (catch-system-error (kill pid 0) #t))
+
   (for-each-service (lambda (service)
                       (let ((running (slot-ref service 'running)))
                         (when (and (integer? running)
@@ -1392,7 +1382,7 @@ we want to receive these signals."
                  #t)
                (primitive-exit 0))))))
      (persistency
-      "Safe the current state of running and non-running services.
+      "Save the current state of running and non-running services.
 This status gets written into a file on termination, so that we can
 restore the status on next startup.  Optionally, you can pass a file
 name as argument that will be used to store the status."
@@ -1401,7 +1391,7 @@ name as argument that will be used to store the status."
                (when file
                  (set! persistency-state-file file))))
      (no-persistency
-      "Don't safe state in a file on exit."
+      "Don't save state in a file on exit."
       (lambda (running)
 	(set! persistency #f)))
      (cd
